# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Accessible tool for statistical analysis of MEG and EEG data.

For the full documentation see <http://eelbrain.readthedocs.io>.



  "The human brain is like an enormous fish. It's flat
   and slimy, and has gills through which it can see."
                                               - Dr. Quat



"""
# import this first for a chance to reverse mne's non-standard logging practice
from . import mne_fixes

from ._config import configure
from ._celltable import Celltable
from ._data_obj import Datalist, Dataset, Var, Factor, Interaction, Model, NDVar, Case, Categorial, Scalar, Sensor, SourceSpace, VolumeSourceSpace, UTS, Space, choose, combine, align, align1, cellname, shuffled_index
from ._experiment import MneExperiment
from ._mne import complete_source_space, labels_from_clusters, morph_source_space, resample_ico_source_space, xhemi
from ._ndvar import Butterworth, concatenate, convolve, correlation_coefficient, cross_correlation, cwt_morlet, dss, filter_data, find_intervals, find_peaks, frequency_response, gaussian, label_operator, maximum, minimum, neighbor_correlation, normalize_in_cells, powerlaw_noise, psd_welch, rename_dim, resample, segment, set_connectivity, set_parc, set_time, set_tmin
from ._ndvar.edge_detector import edge_detector
from ._ndvar.gammatone import gammatone_bank
from ._ndvar.uts import pad
from ._stats.testnd import NDTest, MultiEffectNDTest
from ._trf._boosting import boosting, BoostingResult
from ._trf._predictors import epoch_impulse_predictor, event_impulse_predictor
from ._utils import set_log_level
from ._utils.com import check_for_update

from . import datasets
from . import gui
from . import load
from . import plot
from . import report
from . import save
from . import table
from . import test
from . import testnd

from .fmtxt import Report


__version__ = '0.39.4'



"""Create tables from data-objects"""
# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from itertools import zip_longest
from operator import itemgetter
import re
from typing import Callable, Sequence, Union

import numpy as np

from . import fmtxt
from ._celltable import Celltable
from ._exceptions import KeysMissing
from ._data_obj import (
    CategorialArg, CellArg, IndexArg, ModelArg, NDVarArg, UVArg, VarArg,
    Categorial, Dataset, Factor, Interaction, NDVar, Scalar, UTS,
    Var, ascategorial, as_legal_dataset_key, asndvar, asvar, assub, asuv,
    cellname, combine, isuv)
from ._utils import deprecate_ds_arg


@deprecate_ds_arg
def difference(
        y: Union[NDVar, VarArg, Sequence[Union[NDVar, VarArg]]],
        x: CategorialArg,
        c1: CellArg,
        c0: CellArg,
        match: CategorialArg,
        sub: IndexArg = None,
        data: Dataset = None,
) -> Dataset:
    """Subtract data in one cell from another

    Parameters
    ----------
    y
        One or several variables for which to calculate the difference.
    x
        Model for subtraction, providing categories to compute ``c1 - c0``.
    c1
        Name of the cell in ``x`` that forms the minuend.
    c0
        Name of the cell in ``x`` that is to be subtracted from ``c1``.
    match
        Units over which measurements were repeated. ``c1 - c0`` will be
        calculated separately for each level of ``match`` (e.g. ``"subject"``,
        or ``"subject % condition"``).
    sub
        Only include a subset of the data.
    data
        If a Dataset is specified other arguments can be str instead of
        data-objects and will be retrieved from ``data``.

    Returns
    -------
    diff
        Dataset with the difference between ``c1`` and ``c0`` on ``y``.

    Examples
    --------
    ERP difference wave: assuming a dataset ``data`` with EEG data
    (``data['eeg']``), a variable named ``'condition'`` with levels ``'expected'``
    and ``'unexpected'``, and multiple subjects, the following will generate the
    ``unexpected - expected`` difference waves::

        >>> diff = table.difference('eeg', 'condition', 'unexpected', 'expected',
        ... 'subject', data=data)

    If ``data`` also contains a different factor crossed with ``condition``,
    called ``'word'`` with levels ``'verb'`` abd ``'adjective'``, then separate
    difference waves for verbs and adjectives can be computed with::

        >>> diff = table.difference('eeg', 'condition', 'unexpected', 'expected',
        ... 'subject % word', data=data)

    Given the latter, the difference of the difference waves could be computed
    with::

        >>> diffdiff = table.difference('eeg', 'word', 'verb', 'adjective',
        ... 'subject', data=diff)

    """
    if isinstance(y, (str, Var, NDVar)):
        ys = [y]
    elif len(y) == 0:
        raise ValueError(f'{y=}')
    else:
        ys = y

    out = None
    for yi in ys:
        ct = Celltable(yi, x, match, sub, (c1, c0), data=data)
        if not ct.all_within:
            raise ValueError("Design is not fully balanced")
        if out is None:
            out = Dataset()
            groups = ct.groups[c1]
            if isinstance(groups, Interaction):
                for x_i in groups.base:
                    out.add(x_i)
            else:
                out.add(groups)
        out[ct.y.name] = ct.data[c1] - ct.data[c0]  # use ct.y.name because yi can be expression
    # Transfer other variables in data that are compatible with the rm-structure
    if data is not None:
        out.update(ct._align_ds(data, True, out.keys(), isuv))
    return out


@deprecate_ds_arg
def frequencies(
        y: UVArg,
        x: CategorialArg = None,
        of: CategorialArg = None,
        sub: IndexArg = None,
        data: Dataset = None,
):
    """Calculate frequency of occurrence of the categories in ``y``

    Parameters
    ----------
    y : univariate
        Values whose frequencies are of interest.
    x : categorial
        Optional model defining cells for which frequencies are displayed
        separately.
    of : categorial
        With ``x`` constant within ``of``, only count frequencies for each value
        in ``of`` once. (Compress y and x before calculating frequencies.)
    sub : index
        Only use a subset of the data.
    data : Dataset
        If data is specified, other parameters can be strings naming for
        variables in ``data``.

    Returns
    -------
    freq : Dataset
        Dataset with frequencies.

    Examples
    --------
    A simple sample dataset::

        >>> data = Dataset()
        >>> data['a'] = Factor('aabbcc')
        >>> data['x'] = Factor('xxxyyy')

    Display frequency of a single factor's cells::

        >>> print(table.frequencies('a', data=data))
        cell   n
        --------
        a      2
        b      2
        c      2

    Display frequency of interaction cells::

        >>> print(table.frequencies('a', 'x', data=data))
        x   a   b   c
        -------------
        x   2   1   0
        y   0   1   2

    """
    sub = assub(sub, data)
    y = asuv(y, sub, data, interaction=True)
    if x is not None:
        x = ascategorial(x, sub, data)
    if of is not None:
        of = ascategorial(of, sub, data)
        y = y.aggregate(of)
        if x is not None:
            x = x.aggregate(of)

    name = f"Frequencies of {y.name}" if y.name else "Frequencies"
    if isinstance(y, Var):
        cells = np.unique(y.x)
    else:
        cells = y.cells

    # special case
    if x is None:
        out = Dataset(name=name)
        if isinstance(y, Interaction):
            for i, f in enumerate(y.base):
                out[f.name] = Factor((c[i] for c in x.cells),
                                     random=getattr(f, 'random', False))
        elif isinstance(y, Factor):
            out[y.name or 'y'] = Factor(cells, random=y.random)
        elif isinstance(y, Var):
            out[y.name or 'y'] = Var(cells)
        else:
            raise RuntimeError("y=%r" % (y,))
        n = np.fromiter((np.sum(y == cell) for cell in cells), int, len(cells))
        n_underline = 0
        while (key := 'n' + '_'*n_underline) in out:
            n_underline += 1
        out[key] = Var(n)
        return out

    # header
    if getattr(x, 'name', None):
        name += f' by {x.name}'
    out = Dataset(name=name)

    if isinstance(x, Interaction):
        for i, f in enumerate(x.base):
            random = getattr(f, 'random', False)
            out[f.name] = Factor((c[i] for c in x.cells), random=random)
    else:
        out[x.name] = Factor(x.cells)

    y_idx = {cell: y == cell for cell in cells}
    x_idx = {cell: x == cell for cell in x.cells}
    for y_cell in cells:
        n = (np.sum(np.logical_and(y_idx[y_cell], x_idx[x_cell]))
             for x_cell in x.cells)
        name = as_legal_dataset_key(cellname(y_cell, '_'))
        out[name] = Var(np.fromiter(n, int, len(x.cells)))

    return out


def melt(
        name: str,
        cells: CellArg,
        cell_var_name: str,
        data: Dataset,
        labels: Union[dict, Sequence[str]] = None,
) -> Dataset:
    """
    Restructure a Dataset such that a measured variable is in a single column

    Restructure a Dataset with a certain variable represented in several
    columns into a longer dataset in which the variable is represented in a
    single column along with an identifying variable.

    Additional variables are automatically included.

    Parameters
    ----------
    name
        Name of the variable in the new Dataset.
    cells
        Names of the columns representing the variable in the input Dataset.
        Names can either pe specified explicitly as a sequence of str, or
        implicitly as a str containing '%i' for an integer.
    cell_var_name
        Name of the variable to contain the cell identifier.
    data
        Input data.
    labels
        Labels for the keys in ``cells``. Can be specified either as a
        ``{key: label}`` dictionary or as a list of :class:`str` corresponding
        to ``cells``.

    Examples
    --------
    Simple example data::

        >>> data = Dataset()
        >>> data['y1'] = Var([1, 2, 3])
        >>> data['y2'] = Var([4, 5, 6])
        >>> print(data)
        y1   y2
        -------
        1    4
        2    5
        3    6
        >>> print(table.melt('y', ['y1', 'y2'], 'id', data))
        y   id
        ------
        1   y1
        2   y1
        3   y1
        4   y2
        5   y2
        6   y2

    Additional variables are automatically included::

        >>> data['rm'] = Factor('abc')
        >>> print(data)
        y1   y2   rm
        ------------
        1    4    a
        2    5    b
        3    6    c
        >>> print(table.melt('y', ['y1', 'y2'], 'id', data))
        rm   y   id
        -----------
        a    1   y1
        b    2   y1
        c    3   y1
        a    4   y2
        b    5   y2
        c    6   y2

    """
    # find source cells
    if isinstance(cells, str):
        cell_expression = cells
        cells = []
        cell_values = []
        if '%i' in cell_expression:
            pattern = cell_expression.replace('%i', r'(\d+)')
            for key in data:
                m = re.match(pattern, key)
                if m:
                    cells.append(key)
                    cell_values.append(int(m.group(1)))
        else:
            raise ValueError(f"cells={cells!r}; If specified as string, it needs to contain '%i' as a place-holder for an integer that identifies columns")
    else:
        cell_values = cells

    if labels is None:
        cell_labels = cell_values
    elif isinstance(labels, dict):
        cell_labels = [labels[v] for v in cell_values]
    elif len(labels) != len(cells):
        raise ValueError(f"labels={labels!r}: needs as many entries as there are cells ({len(cells)})")
    else:
        cell_labels = labels

    # melt the Dataset
    keep = [k for k, v in data.items() if isuv(v)]
    if keep:
        out = data[tuple(keep)].tile(len(cells))
    else:
        out = Dataset(info=data.info)
    out[name] = combine(data[cell] for cell in cells)
    out[cell_var_name] = Factor(cell_labels, repeat=data.n_cases)
    return out


def melt_ndvar(
        ndvar: NDVarArg,
        dim: str = None,
        cells: Sequence = None,
        ds: Dataset = None,
        varname: str = None,
        labels: Union[dict, Callable] = None,
) -> Dataset:
    """
    Transform data to long format by converting an NDVar dimension into a variable

    Parameters
    ----------
    ndvar
        The NDVar (or name of the NDVar in ``ds``).
    dim
        The name of the dimension that should be unwrapped (optional if
        ``ndvar`` has only one non-case dimension).
    cells
        The values on ``dim`` that should be included in the output Dataset
        (default is to include all values).
    ds
        Dataset with additional variables that should be included in the long
        table.
    varname
        Name for the transformed ``ndvar`` (default is ``ndvar.name``).
    labels
        Mapping or function to create labels for ``dim`` levels.

    Returns
    -------
    long_table : Dataset
        Dataset in long format.

    See Also
    --------
    cast_to_ndvar

    Examples
    --------
    See :ref:`exa-compare-topographies`.
    """
    ndvar = asndvar(ndvar, data=ds)
    if dim is None:
        if ndvar.ndim == ndvar.has_case + 1:
            dim = ndvar.dims[-1]
            dimname = dim.name
        else:
            raise ValueError(f"The ndvar has more than one possible dimensions, the dim parameter must be one of {ndvar.dimnames[ndvar.has_case:]}")
    else:
        dimname = dim
        dim = ndvar.get_dim(dimname)

    dim_vars = dim._melt_vars()
    if cells is None:
        cells = dim._as_uv()
    elif dim_vars:
        index = dim._array_index(cells)
        dim_vars = {k: v[index] for k, v in dim_vars.items()}

    if callable(labels):
        label = labels
    elif isinstance(labels, dict):
        missing = set(cells).difference(labels)
        if missing:
            raise KeysMissing(missing, 'labels', labels)
        label = itemgetter(labels)
    elif labels is None:
        label = lambda x: x
    else:
        raise TypeError(f"labels={labels!r}")

    if varname is None:
        if ndvar.name is None:
            raise TypeError("Need to provide a name")
        varname = ndvar.name

    if ds is None:
        base_ds = Dataset()
    else:
        uv_keys = tuple(k for k, v in ds.items() if isuv(v))
        base_ds = ds[uv_keys]

    if not ndvar.has_case:
        ndvar = ndvar[np.newaxis]
    dss = []
    for cell in cells:
        ds_ = base_ds.copy()
        ds_[varname] = ndvar.sub(**{dimname: cell})
        ds_[dimname, :] = label(cell)
        dss.append(ds_)
    out = combine(dss)
    out.info['varname'] = varname
    for k, v in dim_vars.items():
        if k not in out:
            out[k] = v.repeat(len(ndvar))
    return out


@deprecate_ds_arg
def cast_to_ndvar(
        y: Union[VarArg, Sequence[VarArg]],
        dim_values: Union[VarArg, Factor],
        match: CategorialArg,
        sub: IndexArg = None,
        data: Dataset = None,
        dim: str = None,
        unit: str = 's',
        name: str = None,
) -> Dataset:
    """Create an NDVar by converting a data column to a dimension
    
    Parameters
    ----------
    y
        Data to be cast.
    dim_values
        Location on the new dimension.
    match
        Indicating rows which belong the the same case in the NDvar.
    sub
        Use a subset of the data.
    data
        Dataset with data for operation.
    dim
        Name for the new dimension. Use ``dim='uts'`` to create :class:`UTS` 
        time dimension from scalar ``dim_values``.
    unit
        Unit for :class:`UTS` dimension (ignored otherwise).
    name
        Name for the new :class:`NDVar` (the default is the name of ``data``).
        
    Returns
    -------
    short_ds
        Copy of ``ds``, aggregated over ``dim_values``, and with an 
        :class:`NDVar` containing the values form ``data`` and a new dimension
        reflecting ``dim_values``. If ``dim_values`` is a Factor, the new 
        dimension is :class:`Categorial`; if ``dim_values`` is a :class:`Var`, 
        it is :class:`Scalar`. The new dimension's name is ``dim``. The only
        exception to this is that when ``dim='uts'``, the new dimension is 
        :class:`UTS` named ``'time'``.

    See Also
    --------
    melt_ndvar
    """
    sub, n = assub(sub, data, return_n=True)
    if isinstance(y, (str, Var)):
        y, n = asvar(y, sub, data, n, return_n=True)
        data_vars = [y]
        names = [name]
    else:
        data_vars = []
        for data_i in y:
            data_var, n = asvar(data_i, sub, data, n, return_n=True)
            data_vars.append(data_var)
        if name is None:
            names = [None for _ in range(len(data_vars))]
        elif isinstance(name, str):
            raise TypeError(f"name={name!r}: single name for multiple variables")
        else:
            names = name
    dim_values, n = asuv(dim_values, sub, data, n, return_n=True)
    match, n = ascategorial(match, sub, data, n, return_n=True)

    # determine NDVar dimension
    if isinstance(dim_values, Factor):
        unique_dim_values = dim_values.cells
        dim = Categorial(dim or dim_values.name, unique_dim_values)
    else:
        unique_dim_values = np.unique(dim_values.x)
        if dim == 'uts':
            diff = np.diff(unique_dim_values)
            unique_diff = np.unique(diff)
            if len(unique_diff) > 1:
                if np.diff(unique_diff).max() > 1e-15:
                    raise NotImplementedError(f"Can't create UTS dimension from data with irregular sampling (detected time-steps of {', '.join(map(str, unique_diff))}")
                tstep = round(unique_diff.mean(), 17)
            else:
                tstep = unique_diff[0]
            dim = UTS(unique_dim_values[0], tstep, len(unique_dim_values), unit)
        else:
            dim = Scalar(dim or dim_values.name, unique_dim_values)

    # find NDVar data
    n_samples = len(dim)
    n_cases = len(match.cells)
    case_indexes = [match == case for case in match.cells]
    samples_indexes = [dim_values == v for v in unique_dim_values]
    xs = [np.empty((n_cases, n_samples)) for _ in data_vars]
    index = None
    for i, case_index in enumerate(case_indexes):
        try:
            for j, sample_index in enumerate(samples_indexes):
                index = np.logical_and(case_index, sample_index, out=index)
                for x, data_var in zip(xs, data_vars):
                    x[i, j] = data_var.x[index]
        except ValueError:
            if not np.any(index):
                raise ValueError(f"Case {match.cells[i]!r} is missing some values")

    # package output dataset
    if data is None:
        out = Dataset()
    else:
        out = data if sub is None else data.sub(sub)
        out = out.aggregate(match, drop_bad=True, count=False)
    for x, data_var, name in zip_longest(xs, data_vars, names):
        out[name or data_var.name] = NDVar(x, ('case', dim))
    return out


@deprecate_ds_arg
def stats(
        y: Var,
        row: CategorialArg,
        col: CategorialArg = None,
        match: CategorialArg = None,
        sub: IndexArg = None,
        fmt: str = '%.4g',
        funcs: Sequence[Union[str, Callable]] = ('mean',),
        data: Dataset = None,
        title: fmtxt.FMTextArg = None,
        caption: fmtxt.FMTextArg = None,
        format: bool = True,
) -> Union[Dataset, fmtxt.Table]:
    """Make a table with statistics

    Parameters
    ----------
    y
        Dependent variable.
    row
        Model specifying rows
    col
        Model specifying columns.
    match
        Identifier for repeated measures data; aggregate within subject before
        computing statistics.
    sub
        Only use part of the data.
    fmt
        How to format values.
    funcs
        A list of statistics functions to show (all functions must take an
        array argument and return a scalar; strings are interpreted as
        :mod:`numpy` functions).
    data
        If a Dataset is provided, ``y``, ``row``, and ``col`` can be strings
        specifying members.
    title
        Table title.
    caption
        Table caption.
    format
        Return a formatted table (instead of a :class:`Dataset`)


    Returns
    -------
    table
        Table with statistics.

    Examples
    --------
    >>> data = datasets.get_uts()
    >>> table.stats('Y', 'A', 'B', data=data)
                B
         -----------------
         b0        b1
    ----------------------
    a0   0.1668    -0.3646
    a1   -0.4897   0.8746

    >>> table.stats('Y', 'A', data=data, funcs=['mean', 'std'])
    Condition   Mean     Std
    --------------------------
    a0          0.6691   1.37
    a1          0.8596   1.192

    """
    sub = assub(sub, data)
    y = asvar(y, sub, data)
    row = ascategorial(row, sub, data)
    if match is not None:
        match = ascategorial(match, sub, data)

    if isinstance(funcs, str):
        funcs = [funcs]
    funcs = [getattr(np, f) if isinstance(f, str) else f for f in funcs]

    if col is None:
        ct = Celltable(y, row, match=match)
        if not format:
            out = Dataset()
            if isinstance(ct.x, Factor):
                out[ct.x.name or 'cell'] = Factor(ct.x.cells)
            elif isinstance(ct.x, Interaction):
                for i, base in enumerate(ct.x.base):
                    out[base.name or f'cell_{i}'] = Factor([cell[i] for cell in ct.x.cells])
            else:
                raise RuntimeError(f"{ct.x=}")

            for func in funcs:
                out[func.__name__] = ct.get_statistic(func)
            return out

        # table header
        n_disp = len(funcs)
        table = fmtxt.Table('l' * (n_disp + 1), title=title, caption=caption)
        table.cell('Condition', 'bf')
        for func in funcs:
            table.cell(func.__name__.capitalize(), 'bf')
        table.midrule()

        # table entries
        for cell in ct.cells:
            data = ct.data[cell]
            table.cell(cell)
            for func in funcs:
                table.cell(fmt % func(data.x))
    else:
        col = ascategorial(col, sub, data)
        ct = Celltable(y, row % col, match=match)

        N = len(col.cells)
        if not format:
            raise NotImplementedError(f"{format=} with col specified")

        table = fmtxt.Table('l' * (N + 1), title=title, caption=caption)

        # table header
        table.cell()
        table.cell(col.name, width=N, just='c')
        table.midrule(span=(2, N + 1))
        table.cell()

        table.cells(*col.cells)
        table.midrule()

        # table body
        fmt_n = fmt.count('%')
        if fmt_n == 1:
            fmt_once = False
        elif len(funcs) == fmt_n:
            fmt_once = True
        else:
            raise ValueError("fmt does not match funcs")

        for Ycell in row.cells:
            table.cell(Ycell)
            for Xcell in col.cells:
                # construct address
                a = ()
                if isinstance(Ycell, tuple):
                    a += Ycell
                else:
                    a += (Ycell,)
                if isinstance(Xcell, tuple):
                    a += Xcell
                else:
                    a += (Xcell,)

                # cell
                data = ct.data[a]
                values = (f(data.x) for f in funcs)
                if fmt_once:
                    txt = fmt % values
                else:
                    txt = ', '.join((fmt % v for v in values))

                table.cell(txt)

    return table


def repmeas(
        y: Union[NDVarArg, ModelArg],
        x: CategorialArg,
        match: CategorialArg,
        sub: IndexArg = None,
        data: Dataset = None,
) -> Dataset:
    """Create a repeated-measures table

    Parameters
    ----------
    y
        Dependent variable (can be model with several dependents).
    x
        Model defining the cells that should be restructured into variables.
    match : categorial
        Model identifying the source of the measurement across repetitions,
        i.e. the model that should be retained.
    sub
        boolean array specifying which values to include (generate e.g.
        with 'sub=T==[1,2]')
    data
        If a Dataset is specified other arguments can be str instead of
        data-objects and will be retrieved from ds.

    Returns
    -------
    rm_table : Dataset
        Repeated measures table. Entries for cells of ``x`` correspond to the
        data in ``y`` on these levels of ``x`` (if cell names are not valid
        Dataset keys they are modified).

    Examples
    --------
    Generate test data in long format::

        >>> ds = Dataset()
        >>> data['y'] = Var([1,2,3,5,6,4])
        >>> data['x'] = Factor('aaabbb')
        >>> data['rm'] = Factor('123231', random=True)
        >>> print(data)
        y   x   rm
        ----------
        1   a   1
        2   a   2
        3   a   3
        5   b   2
        6   b   3
        4   b   1

    Compute difference between two conditions::

        >>> ds_rm = table.repmeas('y', 'x', 'rm', data=data)
        >>> print(ds_rm)
        rm   a   b
        ----------
        1    1   4
        2    2   5
        3    3   6
        >>> ds_rm['difference'] = ds_rm.eval("b - a")
        >>> print(ds_rm)
        rm   a   b   difference
        -----------------------
        1    1   4   3
        2    2   5   3
        3    3   6   3

    """
    ct = Celltable(y, x, match, sub, data=data)
    if not ct.all_within:
        raise ValueError("Incomplete data")

    out = Dataset()
    x_ = ct.groups[ct.x.cells[0]]
    if isinstance(x_, Interaction):
        for f in x_.base:
            out.add(f)
    else:
        out[ct.match.name] = x_

    for cell in ct.x.cells:
        if len(ct.data[cell]):  # for models with empty cells
            key = as_legal_dataset_key(cellname(cell, '_'))
            out[key] = ct.data[cell]

    # Transfer other variables in ds that are compatible with the rm-structure
    if data is not None:
        out.update(ct._align_ds(data, True, out.keys(), isuv))

    return out



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from typing import Any, Literal, Union

import matplotlib.axes
import matplotlib.patches

from .._stats import test


def figure_outline(color='k', figure=None, **kwargs):
    """Draw the outline of the current figure

    Mainly for fine-tuning the figure layout in Jupyter, which crops
    the display area based on figure elements rather than actual figure size.
    """
    if figure is None:
        from matplotlib import pyplot
        figure = pyplot.gcf()
    kwargs.setdefault('fc', 'none')
    artist = matplotlib.patches.Rectangle((0, 0), 1, 1, ec=color, **kwargs)
    figure.add_artist(artist)


def mark_difference(
        x1: float,
        x2: float,
        y: float,
        mark: Union[float, str] = None,
        dy: float = 0,
        color: Any = None,
        nudge: Union[bool, float] = None,
        location: Literal['top', 'bottom', 'left', 'right'] = 'top',
        ax: matplotlib.axes.Axes = None,
        line_args: dict = None,
        **text_args,
):
    """Mark a pair of categories with a line and a label

    Parameters
    ----------
    x1
        Location on category axis.
    x2
        Second location on category axis.
    y
        Level above which to plot the bar.
    mark
        Text label, or p-value to automatically determine the label and
        ``color``.
    dy
        Length of vertical ticks on each side of the bar (offsets the
        location of the bar itself to ``y + dy``).
    color
        Color for bar and ``label``.
    nudge
        Nudge the edges of the bar inwards to allow multiple bars
        side-by-side on the same level of ``y``.
    location
        Location relative to plots.
    ax
        Axes to which to plot (default is ``pyplot.gca()``).
    line_args
        Additional parameters for :func:`matplotlib.pyplot.plot`.
    ...
        All other parameters are used to plot the text label with
        :meth:`matplotlib.axes.Axes.text`.
    """
    if ax is None:
        from matplotlib import pyplot
        ax = pyplot.gca()

    if isinstance(mark, str):
        label = mark
    elif mark is None:
        label = ''
    else:
        p = mark
        n_stars = test._n_stars(p)
        if not n_stars:
            return
        if color is None:
            color = ('#FFCC00', '#FF6600', '#FF3300')[n_stars - 1]
        label = '*' * n_stars
        text_args.setdefault('size', matplotlib.rcParams['font.size'] * 1.5)

    if color is None:
        color = 'k'

    if x1 > x2:
        x1, x2 = x2, x1

    if nudge:
        if nudge is True:
            nudge = 0.025
        x1 += nudge
        x2 -= nudge

    if location in ('bottom', 'left'):
        dy = -dy

    x_text = (x1 + x2) / 2
    if dy:
        xs, ys = [x1, x1, x2, x2], [y, y + dy, y + dy, y]
        y_text = y + dy
    else:
        xs, ys = [x1, x2], [y, y]
        y_text = y

    if location in ('right', 'left'):
        xs, ys = ys, xs
        x_text, y_text = y_text, x_text

    if x1 != x2:
        line_args_ = {'color': color, 'clip_on': False}
        if line_args:
            line_args_.update(line_args)
        ax.plot(xs, ys, **line_args_)
    if label:
        rotation = {'top': 0, 'left': 90, 'bottom': 180, 'right': 270}[location]
        args = {'ha': 'center', 'va': 'center', 'rotation': rotation, 'clip_on': False, **text_args}
        ax.text(x_text, y_text, label, color=color, **args)



import traceback
import warnings
import sys


def warn_with_traceback():
    "Replace ``warnings.showwarning``"
    def warn(message, category, filename, lineno, file=None, line=None):
        """Print tracebacks for warnings (see https://stackoverflow.com/a/22376126)"""
        log = file if hasattr(file, 'write') else sys.stderr
        traceback.print_stack(file=log)
        log.write(warnings.formatwarning(message, category, filename, lineno, line))

    warnings.showwarning = warn



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from numpy.testing import assert_array_equal
import numpy as np
import pingouin
import pytest
import scipy.stats

from eelbrain import datasets, test
from eelbrain.fmtxt import asfmtext
from eelbrain._stats import test as _test


def test_correlations():
    "Test test.correlations()"
    ds = datasets.get_uv()

    res = test.correlations('fltvar', 'fltvar2', ds=ds)
    print(res)
    assert str(res[2][0]).strip() == '.398'
    res = test.correlations('fltvar', 'fltvar2', ds=ds, asds=True)
    assert res[0, 'r'] == pytest.approx(.398, abs=1e-3)
    res = test.Correlation('fltvar', 'fltvar2', data=ds)
    assert res.r == pytest.approx(.398, abs=1e-3)

    res = test.correlations('fltvar', 'fltvar2', 'A', ds=ds)
    print(res)
    assert str(res[2][0]).strip() == 'a1'
    assert str(res[2][1]).strip() == '-.149'
    assert str(res[3][1]).strip() == '.740'
    res = test.correlations('fltvar', 'fltvar2', 'A', ds=ds, asds=True)
    assert res[0, 'A'] == 'a1'
    assert res[0, 'r'] == pytest.approx(-0.149, abs=1e-3)
    assert res[1, 'r'] == pytest.approx(.740, abs=1e-3)
    res = test.Correlation('fltvar', 'fltvar2', "A == 'a1'", ds)
    assert res.r == pytest.approx(-0.149, abs=1e-3)

    res = test.correlations('fltvar', 'fltvar2', 'A%B', ds=ds)
    print(res)
    assert str(res[2][2]).strip() == '-.276'
    res = test.correlations('fltvar', 'fltvar2', 'A%B', ds=ds, asds=True)
    assert res[0, 'r'] == pytest.approx(-0.276, abs=1e-3)

    res = test.correlations('fltvar', ('fltvar2', 'intvar'), 'A%B', ds=ds)
    print(res)
    assert str(res[2][1]).strip() == 'a1'
    assert str(res[2][2]).strip() == 'b1'
    assert str(res[2][3]).strip() == '-.276'
    res = test.correlations('fltvar', ('fltvar2', 'intvar'), 'A%B', ds=ds, asds=True)
    assert res[0, 'r'] == pytest.approx(-0.276, abs=1e-3)
    res = test.Correlation('fltvar', 'intvar', "(A=='a1')&(B=='b1')", ds)
    assert res.r == pytest.approx(0.315, abs=1e-3)

    # pairwise correlation
    doc = test.pairwise_correlations(['intvar', 'fltvar', 'fltvar2'], data=ds)


def test_mann_whitney():
    ds = datasets.get_uv()

    ds_agg = ds.aggregate('A % rm', drop_bad=True)
    n = ds_agg.n_cases // 2
    a, b = ds_agg[:n, 'fltvar'], ds_agg[n:, 'fltvar']
    u, p = scipy.stats.mannwhitneyu(a.x, b.x, alternative='two-sided')

    res = test.MannWhitneyU('fltvar', 'A', 'a1', 'a2', 'rm', data=ds)
    assert res.u == u
    assert res.p == p

    res = test.MannWhitneyU(a, b)
    assert res.u == u
    assert res.p == p


def test_star():
    "Test the star function"
    assert_array_equal(_test.star([0.1, 0.04, 0.01], int), [0, 1, 2])
    assert_array_equal(_test.star([0.001, 0.04, 0.1], int), [3, 1, 0])


def test_ttest():
    """Test univariate t-test functions"""
    ds = datasets.get_uv()

    print(test.ttest('fltvar', data=ds))
    print(test.ttest('fltvar', 'A', data=ds))
    print(test.ttest('fltvar', 'A%B', data=ds))
    print(test.ttest('fltvar', 'A', match='rm', data=ds))
    print(test.ttest('fltvar', 'A', 'a1', match='rm', data=ds))
    print(test.ttest('fltvar', 'A%B', ('a1', 'b1'), match='rm', data=ds))

    # Prepare data for scipy
    a1_index = ds.eval("A == 'a1'")
    a2_index = ds.eval("A == 'a2'")
    b1_index = ds.eval("B == 'b1'")
    a1_in_b1_index = np.logical_and(a1_index, b1_index)
    a2_in_b1_index = np.logical_and(a2_index, b1_index)
    a1 = ds[a1_index, 'fltvar'].x
    a2 = ds[a2_index, 'fltvar'].x
    a1_in_b1 = ds[a1_in_b1_index, 'fltvar'].x
    a2_in_b1 = ds[a2_in_b1_index, 'fltvar'].x

    # TTest1Samp
    standard = pingouin.ttest(ds['fltvar'], 0)
    res = test.TTestOneSample('fltvar', data=ds)
    t, p = scipy.stats.ttest_1samp(ds['fltvar'], 0)
    assert res.t == pytest.approx(t, 10)
    assert res.p == pytest.approx(p, 10)
    assert res.d == pytest.approx(standard['cohen-d'][0], 10)
    assert str(res.full) == 'M = 0.40, SD = 1.20, t(79) = 2.96, p = .004'
    res = test.TTestOneSample('fltvar', data=ds, tail=1)
    assert res.t == pytest.approx(t, 10)
    assert res.p == pytest.approx(p / 2., 10)
    assert res.d == pytest.approx(standard['cohen-d'][0], 10)
    assert str(res.full) == 'M = 0.40, SD = 1.20, t(79) = 2.96, p = .002'

    # TTestIndependent
    cohens_d = pingouin.compute_effsize(a1, a2)
    res = test.TTestIndependent('fltvar', 'A', 'a1', 'a2', data=ds)
    t, p = scipy.stats.ttest_ind(ds[a1_index, 'fltvar'], ds[a2_index, 'fltvar'])
    assert res.t == pytest.approx(t, 10)
    assert res.p == pytest.approx(p, 10)
    assert str(res.full) == 'a1: M = 1.00, SD = 1.02; a2: M = -0.20, SD = 1.05; t(78) = 5.10, p < .001'
    assert res.d == cohens_d

    # TTestRelated
    res = test.TTestRelated('fltvar', 'A', 'a1', 'a2', 'rm', "B=='b1'", ds)
    standard = pingouin.ttest(a1_in_b1, a2_in_b1)
    difference = a1_in_b1 - a2_in_b1
    t, p = scipy.stats.ttest_rel(a1_in_b1, a2_in_b1)
    assert_array_equal(res.difference.x, difference)
    assert res.df == len(a1_in_b1) - 1
    assert res.tail == 0
    assert res.t == pytest.approx(t)
    assert res.p == pytest.approx(p)
    assert res.d == pytest.approx(standard['cohen-d'][0], 10)
    print(res)
    print(asfmtext(res))
    assert str(res.full) == 'a1: M = 0.90; a2: M = -0.06; difference: M = 0.96, SD = 1.65, t(19) = 2.53, p = .021'

    res = test.TTestRelated('fltvar', 'A', 'a1', 'a2', 'rm', "B=='b1'", ds, 1)
    assert_array_equal(res.difference.x, difference)
    assert res.df == len(a1_in_b1) - 1
    assert res.tail == 1
    assert res.t == pytest.approx(t)
    assert res.p == pytest.approx(p / 2)
    print(res)
    print(asfmtext(res))

    res = test.TTestRelated('fltvar', 'A', 'a2', 'a1', 'rm', "B=='b1'", ds, 1)
    assert_array_equal(res.difference.x, -difference)
    assert res.df == len(a1_in_b1) - 1
    assert res.tail == 1
    assert res.t == pytest.approx(-t)
    assert res.p == pytest.approx(1 - p / 2)

    res = test.TTestRelated('fltvar', 'A', 'a1', 'a2', 'rm', "B=='b1'", ds, -1)
    assert_array_equal(res.difference.x, difference)
    assert res.df == len(a1_in_b1) - 1
    assert res.tail == -1
    assert res.t == pytest.approx(t)
    assert res.p == pytest.approx(1 - p / 2)
    print(res)
    print(asfmtext(res))
    # alternative argspec
    a1_in_b1 = ds.eval("fltvar[(B == 'b1') & (A == 'a1')]")
    a2_in_b1 = ds.eval("fltvar[(B == 'b1') & (A == 'a2')]")
    res_alt = test.TTestRelated(a1_in_b1, a2_in_b1, tail=-1)
    print(res_alt)
    assert res_alt.t == res.t
    assert res_alt.p == res.p


def test_wilcoxon():
    ds = datasets.get_uv()

    ds_agg = ds.aggregate('A % rm', drop_bad=True)
    n = ds_agg.n_cases // 2
    w, p = scipy.stats.wilcoxon(ds_agg[:n, 'fltvar'].x, ds_agg[n:, 'fltvar'].x, alternative='two-sided')

    res = test.WilcoxonSignedRank('fltvar', 'A', 'a1', 'a2', 'rm', data=ds)
    assert res.w == w
    assert res.p == p



"Function to fix old data files from NYU/KIT system"
import numpy as np


def fix_ptb_eeg_events(raw):
    """Fix events from a vmrk file recorded with psychtoolbox/stim tracker

    Parameters
    ----------
    raw : RawBrainVision
        MNE-Python object with psychtoolbox events.
    """
    events = raw.get_brainvision_events()

    if not np.all(events[:, 1] == 1):
        err = ("Not KIT psychtoolbox input data (not all durations are 1)")
        raise ValueError(err)

    # invert trigger codes
    events[:, 2] = np.invert(events[:, 2].astype(np.uint8))

    # extend duration until next tigger start
    events[:-1, 1] = np.diff(events[:, 0])

    # remove 0 events
    idx = np.nonzero(events[:, 2])[0]
    events = events[idx]

    raw.set_brainvision_events(events)



"""NSActivityOptions

https://developer.apple.com/library/content/documentation/Performance/Conceptual/power_efficiency_guidelines_osx/PrioritizeWorkAtTheAppLevel.html

License
-------
Incorporates code from `appnope <https://github.com/minrk/appnope>`_ under
2-clause BSD, copyright (c) 2013, Min Ragan-Kelley

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
# cf. https://github.com/ipython/ipython/tree/master/IPython/terminal/pt_inputhooks/osx.py
import ctypes.util

objc = ctypes.cdll.LoadLibrary(ctypes.util.find_library('objc'))

void_p = ctypes.c_void_p
ull = ctypes.c_uint64

objc.objc_getClass.restype = void_p
objc.sel_registerName.restype = void_p
objc.objc_msgSend.restype = void_p
objc.objc_msgSend.argtypes = [void_p, void_p]

msg = objc.objc_msgSend
n = objc.sel_registerName
c = objc.objc_getClass

# constants from Foundation
NSActivityIdleDisplaySleepDisabled = (1 << 40)
NSActivityIdleSystemSleepDisabled = (1 << 20)
NSActivitySuddenTerminationDisabled = (1 << 14)
NSActivityAutomaticTerminationDisabled = (1 << 15)
NSActivityUserInitiated = (0x00FFFFFF | NSActivityIdleSystemSleepDisabled)
NSActivityUserInitiatedAllowingIdleSystemSleep = (NSActivityUserInitiated & ~NSActivityIdleSystemSleepDisabled)
NSActivityBackground = 0x000000FF
NSActivityLatencyCritical = 0xFF00000000


def _utf8(s):
    return s if isinstance(s, bytes) else s.encode('utf8')


def beginActivityWithOptions(options, reason=""):
    """Begin an activity using the given options and reason

    Returns
    -------
    activity : int
        Identifier for the activity.

    Notes
    -----
    Indicate completion of the activity by calling :func:`end_activity`,
    passing the returned identifier as the argument.
    """
    reason = msg(c(b'NSString'), n(b'stringWithUTF8String:'), _utf8(reason))
    info = msg(c(b'NSProcessInfo'), n(b'processInfo'))
    return msg(info, n(b'beginActivityWithOptions:reason:'), ull(options),
               void_p(reason))


def end_activity(activity):
    """end a process activity assertion

    Parameters
    ----------
    activity : int
        Identifier for the activity, returned by
        :func:`beginActivityWithOptions`.
    """
    info = msg(c(b'NSProcessInfo'), n(b'processInfo'))
    msg(info, n(b'endActivity:'), void_p(activity))


def begin_activity(options=NSActivityUserInitiated, reason='Eelbrain'):
    """Disable macOS system sleep"""
    return beginActivityWithOptions(options, reason)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
import pytest

from eelbrain import datasets, plot
from eelbrain.testing import hide_plots


def test_generate_colors():
    "Test functions for assigning colors to cells"
    ds = datasets.get_uv()
    cells_1 = ('A', 'B')
    cells_2 = ('a', 'b', 'c')

    colors = plot.colors_for_oneway(cells_1)
    assert len(colors) == len(cells_1)

    colors = plot.colors_for_twoway(cells_1, cells_2)
    assert len(colors) == len(cells_1) * len(cells_2)

    colors = plot.colors_for_twoway(cells_2, cells_1)
    assert len(colors) == len(cells_1) * len(cells_2)

    # colors_for_categorial()
    f = ds['A']
    colors = plot.colors_for_categorial(f)
    for cell in f.cells:
        assert cell in colors

    i = ds.eval("A%B")
    colors = plot.colors_for_categorial(i)
    for cell in i.cells:
        assert cell in colors

    i = ds.eval("A%B%rm")
    colors = plot.colors_for_categorial(i)
    for cell in i.cells:
        assert cell in colors

    with pytest.raises(TypeError):
        plot.colors_for_categorial("A%B")


def test_generate_colormaps():
    cmap = plot.two_step_colormap('black', 'red', name='oneside')
    assert not cmap.symmetric
    cmap = plot.two_step_colormap('black', 'red', 'transparent', 'blue', 'black', name='red-blue')
    assert cmap.symmetric
    cmap = plot.two_step_colormap('black', (1, 0, 0.3), 'transparent', (0.3, 0, 1), 'black', name='red-blue-2')
    assert cmap.symmetric


def ticks(p):
    return [t.get_text() for t in p._colorbar.ax.get_xticklabels()]


@hide_plots
def test_plot_colorbar():
    "Test plot.ColorBar()"
    p = plot.ColorBar('jet', -1, 1)
    p.close()
    p = plot.ColorBar('jet', -1, 1, ticks=[-1, 0, 1])
    assert ticks(p) == ['−1', '0', '1']
    p = plot.ColorBar('jet', -1, 1, orientation='vertical')
    p.close()
    p = plot.ColorBar('jet', -1, 1, label_position='top')
    p.close()
    p = plot.ColorBar('jet', -1, 1, orientation='vertical', label_position='right')
    p.close()
    p = plot.ColorBar('xpolar-a', -3, 3, clipmin=0, unit='t', ticks=[0, 2, 3])
    assert ticks(p) == ['0', '2', '3']
    p.close()

    p = plot.ColorBar('xpolar-a', -3e-6, 3e-6, unit='µV')
    assert ticks(p) == ['-3', '-1.5', '0', '1.5', '3']
    assert p._colorbar.ax.get_xlabel() == 'µV'

    # soft-thresholded colormap
    cmap = plot.soft_threshold_colormap('xpolar-a', 0.2, 2.0)
    p = plot.ColorBar(cmap)
    p.close()
    cmap = plot.soft_threshold_colormap('xpolar-a', 1, 3)
    p = plot.ColorBar(cmap, clipmin=0, unit='t', ticks=[0, 2, 3])
    assert ticks(p) == ['0', '2', '3']
    p.close()
    cmap = plot.soft_threshold_colormap('xpolar-a', 0.5, 3)
    p = plot.ColorBar(cmap, clipmin=0, unit='t', ticks=[0, 2, 3])
    assert ticks(p) == ['0', '2', '3']
    p.close()


@hide_plots
def test_plot_colors():
    "Test plotting color schemes"
    cells_1 = ('A', 'B')
    cells_2 = ('a', 'b', 'c')
    colors = plot.colors_for_oneway(cells_1)
    colors2way = plot.colors_for_twoway(cells_1, cells_2)
    labels = {k: f'Long label {k}' for k in cells_2}

    p = plot.ColorList(colors)
    assert p._figtitle is None
    w0, h0 = p.figure.get_size_inches()
    p.close()

    p = plot.ColorList(colors, labels={'A': 'A'*50, 'B': 'Bbb'})
    w, h = p.figure.get_size_inches()
    assert h == h0
    assert w > w0
    p.close()

    p = plot.ColorList(colors2way)
    p.close()
    p = plot.ColorGrid(cells_1, cells_2, colors2way)
    p.close()
    p = plot.ColorGrid(cells_2, cells_1, colors2way)
    p.close()
    p = plot.ColorGrid(cells_1, cells_2, colors2way, labels=labels, w=1.2, h=1.5)
    p.close()



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
# cython: language_level=3, boundscheck=False, wraparound=False, cdivision=True
from libc.math cimport fabs
from cython.parallel import prange
from libc.stdlib cimport malloc, free
import numpy as np

cimport numpy as np

ctypedef np.int8_t INT8
ctypedef np.int64_t INT64
ctypedef np.float64_t FLOAT64


cdef double inf = float('inf')


cdef double square(double x) nogil:
    # cf. https://github.com/scikit-image/scikit-image/issues/3026
    return x * x


cdef void l1_for_delta(
        FLOAT64 * y_error,
        FLOAT64 [:] x,
        double x_pad,  # pad x outside valid convolution area
        INT64 [:,:] indexes,  # training segment indexes
        double delta,
        int shift,  # TRF element offset
        double* e_add,
        double* e_sub,
    ) nogil:
    cdef:
        double d
        Py_ssize_t i, seg_i, seg_start, seg_stop, conv_start, conv_stop

    e_add[0] = 0.
    e_sub[0] = 0.

    for seg_i in range(indexes.shape[0]):
        seg_start = indexes[seg_i, 0]
        if seg_start == -1:
            break
        seg_stop = indexes[seg_i, 1]
        # determine valid convolution segment
        conv_start = seg_start
        conv_stop = seg_stop
        if shift > 0:
            conv_start += shift
        elif shift < 0:
            conv_stop += shift
        # padding
        d = delta * x_pad
        # pre-
        for i in range(seg_start, conv_start):
            e_add[0] += fabs(y_error[i] - d)
            e_sub[0] += fabs(y_error[i] + d)
        # post-
        for i in range(conv_stop, seg_stop):
            e_add[0] += fabs(y_error[i] - d)
            e_sub[0] += fabs(y_error[i] + d)
        # valid segment
        for i in range(conv_start, conv_stop):
            d = delta * x[i - shift]
            e_add[0] += fabs(y_error[i] - d)
            e_sub[0] += fabs(y_error[i] + d)


cdef void l2_for_delta(
        FLOAT64 * y_error,
        FLOAT64 [:] x,
        double x_pad,  # pad x outside valid convolution area
        INT64 [:,:] indexes,  # training segment indexes
        double delta,
        int shift,
        double* e_add,
        double* e_sub,
    ) nogil:
    cdef:
        double d
        Py_ssize_t i, seg_i, seg_start, seg_stop, conv_start, conv_stop

    e_add[0] = 0.
    e_sub[0] = 0.

    for seg_i in range(indexes.shape[0]):
        seg_start = indexes[seg_i, 0]
        if seg_start == -1:
            break
        seg_stop = indexes[seg_i, 1]
        # determine valid convolution segment
        conv_start = seg_start
        conv_stop = seg_stop
        if shift > 0:
            conv_start += shift
        elif shift < 0:
            conv_stop += shift
        # padding
        d = delta * x_pad
        # pre-
        for i in range(seg_start, conv_start):
            e_add[0] += square(y_error[i] - d)
            e_sub[0] += square(y_error[i] + d)
        # post-
        for i in range(conv_stop, seg_stop):
            e_add[0] += square(y_error[i] - d)
            e_sub[0] += square(y_error[i] + d)
        # part of the segment that is affected
        for i in range(conv_start, conv_stop):
            d = delta * x[i - shift]
            e_add[0] += square(y_error[i] - d)
            e_sub[0] += square(y_error[i] + d)


def error_for_indexes(
        FLOAT64 [:] x,
        INT64[:,:] indexes,  # (n_segments, 2)
        int error,  # 1 --> l1; 2 --> l2
):
    # cdef FLOAT64 * x_ptr = <FLOAT64*> &x[0]
    return error_for_indexes_c(&x[0], indexes, error)


cdef double error_for_indexes_c(
        FLOAT64 * x,
        INT64[:,:] indexes,  # (n_segments, 2)
        int error,  # 1 --> l1; 2 --> l2
) nogil:
    cdef:
        Py_ssize_t seg_i, i
        double out = 0

    if error == 1:
        for seg_i in range(indexes.shape[0]):
            if indexes[seg_i, 0] == -1:
                break
            for i in range(indexes[seg_i, 0], indexes[seg_i, 1]):
                out += fabs(x[i])
    else:
        for seg_i in range(indexes.shape[0]):
            if indexes[seg_i, 0] == -1:
                break
            for i in range(indexes[seg_i, 0], indexes[seg_i, 1]):
                out += x[i] * x[i]
    return out


def boosting_runs(
        FLOAT64[:,:] y,  # (n_y, n_times)
        FLOAT64[:,:] x,  # (n_x, n_times)
        FLOAT64 [:] x_pads,  # (n_x,)
        INT64[:,:,:] split_train,
        INT64[:,:,:] split_validate,
        INT64[:,:,:] split_train_and_validate,
        INT64 [:] i_start_by_x,  # (n_x,) kernel start index
        INT64 [:] i_stop_by_x, # (n_x,) kernel stop index
        double delta,
        double mindelta,
        int error,
        int selective_stopping,
        int num_threads,
):
    """Estimate multiple filters with boosting"""
    cdef:
        Py_ssize_t i_start = np.min(i_start_by_x)
        Py_ssize_t n_y = len(y)
        Py_ssize_t n_x = len(x)
        Py_ssize_t n_splits = len(split_train)
        Py_ssize_t n_total = n_splits * n_y
        Py_ssize_t n_times_h = np.max(i_stop_by_x) - i_start
        FLOAT64[:,:,:,:] hs = np.empty((n_splits, n_y, n_x, n_times_h))
        INT8[:,:] hs_failed = np.zeros((n_splits, n_y), 'int8')
        Py_ssize_t i, i_y, i_split
        BoostingRunResult *result

    for i in prange(n_total, nogil=True, schedule='guided', num_threads=num_threads):
        i_y = i // n_splits
        i_split = i % n_splits
        result = boosting_run(y[i_y], x, x_pads, hs[i_split, i_y], split_train[i_split], split_validate[i_split], split_train_and_validate[i_split], i_start_by_x, i_stop_by_x, delta, mindelta, error, selective_stopping, i_start, n_times_h)
        hs_failed[i_split, i_y] = result.failed
        free_history(result)
    return hs.base, np.asarray(hs_failed, 'bool')


ctypedef struct BoostingStep:
    Py_ssize_t i_step
    Py_ssize_t i_stim
    Py_ssize_t i_time
    double delta
    double e_test
    double e_train
    BoostingStep *previous


ctypedef struct BoostingRunResult:
    int failed
    BoostingStep *history


cdef BoostingRunResult * boosting_run_result(int failed, BoostingStep *history) nogil:
    result = <BoostingRunResult*> malloc(sizeof(BoostingRunResult))
    result.failed = failed
    result.history = history
    return result


cdef void free_history(
        BoostingRunResult *result,
) nogil:
    cdef:
        BoostingStep *step
        BoostingStep *step_i

    step = result.history
    while step.previous != NULL:
        step_i = step.previous
        free(step)
        step = step_i
    free(result)


cdef BoostingRunResult * boosting_run(
        FLOAT64 [:] y,  # (n_times,)
        FLOAT64 [:,:] x,  # (n_x, n_times)
        FLOAT64 [:] x_pads,  # (n_x,)
        FLOAT64 [:,:] h,  # (n_x, n_times_h)
        INT64[:,:] split_train,  # Training data index
        INT64[:,:] split_validate,  # Validation data index
        INT64[:,:] split_train_and_validate,  # Training and validation data index
        INT64 [:] i_start_by_x,  # (n_x,) kernel start index
        INT64 [:] i_stop_by_x, # (n_x,) kernel stop index
        double delta,
        double mindelta,
        int error,
        int selective_stopping,
        Py_ssize_t i_start,
        Py_ssize_t n_times_h,
) nogil:
    cdef:
        int out
        Py_ssize_t n_x = x.shape[0]
        Py_ssize_t n_x_active = n_x
        Py_ssize_t n_times = x.shape[1]
        BoostingStep *step
        BoostingStep *step_j
        BoostingStep *history
        # buffers
        FLOAT64 * y_error = <FLOAT64*> malloc(sizeof(FLOAT64) * n_times)
        INT8 * x_active = <INT8*> malloc(sizeof(INT8) * n_x)
        # history
        double best_test_error = inf
        Py_ssize_t best_step_i = -1
        int n_bad, undo
        Py_ssize_t i_step, i

    # initialize buffers
    h[...] = 0
    for i in range(n_x):
        x_active[i] = 1
    for i in range(n_times):
        y_error[i] = y[i]

    # first step
    step = <BoostingStep*> malloc(sizeof(BoostingStep))
    step.previous = NULL
    step.i_step = -1
    step.i_stim = -1
    step.i_time = -1
    step.delta = 0
    step.e_train = error_for_indexes_c(y_error, split_train, error)
    history = step

    # pre-assign iterators
    for i_step in range(999999):
        # evaluate current h
        e_test = error_for_indexes_c(y_error, split_validate, error)
        step.e_test = e_test

        # evaluate stopping conditions
        if e_test < best_test_error:
            # print(' ', e_test, '<', best_test_error)
            best_test_error = e_test
            best_step_i = step.i_step
        elif e_test > step.previous[0].e_test and i_step >= 2:
            step_j = step.previous
            if selective_stopping:
                undo = 1
                if selective_stopping > 1:
                    # only stop if the predictor overfits n times without intermittent improvement
                    n_bad = 1
                    while True:
                        undo += 1
                        if step_j.i_stim == step.i_stim:
                            if step_j.e_test > step_j.previous[0].e_test:
                                # the same stimulus caused an error increase
                                n_bad += 1
                                if n_bad == selective_stopping:
                                    break
                            else:
                                undo = 0
                                break
                        step_j = step_j.previous
                        if step_j == NULL or step_j.e_test > e_test:
                            undo = 0
                            break  # the error improved

                if undo:
                    # disable predictor
                    x_active[step.i_stim] = 0
                    n_x_active -= 1
                    if n_x_active == 0:
                        break
                    # revert changes
                    for i in range(undo):
                        h[step.i_stim, step.i_time] -= step.delta
                        update_error(y_error, x[step.i_stim], x_pads[step.i_stim], split_train_and_validate, -step.delta, step.i_time + i_start)
                        step_j = step.previous
                        free(step)
                        step = step_j
                    history = step
            # Basic
            # -----
            # stop the iteration if all the following requirements are met
            # 1. more than 10 iterations are done
            # 2. The testing error in the latest iteration is higher than that in
            #    the previous two iterations
            elif i_step > 10 and e_test > step_j.previous[0].e_test:
                # print("error(test) not improving in 2 steps")
                break

        # generate possible movements -> training error
        step = <BoostingStep *> malloc(sizeof(BoostingStep))
        step.previous = history
        step.i_step = i_step
        generate_options(step, y_error, x, x_pads, x_active, split_train, i_start, i_start_by_x, i_stop_by_x, error, delta, n_x)
        history = step

        # If no improvements can be found reduce delta
        if step.e_train > step.previous.e_train:
            delta *= 0.5
            step.delta = 0
            if delta >= mindelta:
                step.i_stim = -1
                step.i_time = -1
                continue
            else:
                # print("No improvement possible for training data")
                break

        # abort if we're moving in circles
        if step.delta == -step.previous.delta and step.i_stim == step.previous.i_stim and step.i_time == step.previous.i_time:
            step.delta = 0
            # print("Moving in circles")
            break

        # update h with best movement
        h[step.i_stim, step.i_time] += step.delta
        update_error(y_error, x[step.i_stim], x_pads[step.i_stim], split_train_and_validate, step.delta, step.i_time + i_start)
    else:
        with gil:
            raise RuntimeError("Boosting: maximum number of iterations exceeded")

    free(y_error)

    # reverse changes after best iteration
    if best_step_i > -1:
        while step.i_step > best_step_i:
            if step.delta:
                h[step.i_stim, step.i_time] -= step.delta
            step = step.previous
        return boosting_run_result(0, history)
    else:
        # print(' failed')
        return boosting_run_result(1, history)


cdef void generate_options(
        BoostingStep * step,
        FLOAT64 * y_error,
        FLOAT64 [:,:] x,  # (n_x, n_times)
        FLOAT64 [:] x_pads,  # (n_x,)
        INT8 * x_active,  # for each predictor whether it is still used
        INT64 [:,:] indexes,  # training segment indexes
        int i_start,  # kernel start index (time axis offset)
        INT64 [:] i_start_by_x,  # (n_x,) kernel start index
        INT64 [:] i_stop_by_x, # (n_x,) kernel stop index
        int error,  # ID of the error function (l1/l2)
        double delta,
        Py_ssize_t n_x,
    ) nogil:
    cdef:
        double e_add, e_sub, e_new, x_pad
        Py_ssize_t i_stim, i_time, new_sign
        FLOAT64 [:] x_stim

    step.e_train = inf
    for i_stim in range(n_x):
        if x_active[i_stim] == 0:
            continue
        x_stim = x[i_stim]
        x_pad = x_pads[i_stim]
        for i_time in range(i_start_by_x[i_stim], i_stop_by_x[i_stim]):
            # +/- delta
            if error == 1:
                l1_for_delta(y_error, x_stim, x_pad, indexes, delta, i_time, &e_add, &e_sub)
            else:
                l2_for_delta(y_error, x_stim, x_pad, indexes, delta, i_time, &e_add, &e_sub)

            i_time -= i_start
            if e_add > e_sub:
                e_new = e_sub
                new_sign = -1
            else:
                e_new = e_add
                new_sign = 1

            # find smallest error
            if e_new < step.e_train:
                step.e_train = e_new
                step.i_stim = i_stim
                step.i_time = i_time
                step.delta = delta * new_sign


cdef void update_error(
        FLOAT64 * y_error,
        FLOAT64 [:] x,
        double x_pad,  # pad x outside valid convolution area
        INT64 [:,:] indexes,  # segment indexes
        double delta,
        Py_ssize_t shift,
    ) nogil:
    cdef:
        Py_ssize_t i, seg_i, seg_start, seg_stop, conv_start, conv_stop

    for seg_i in range(indexes.shape[0]):
        seg_start = indexes[seg_i, 0]
        if seg_start == -1:
            break
        seg_stop = indexes[seg_i, 1]
        conv_start = seg_start
        conv_stop = seg_stop
        if shift > 0:
            conv_start += shift
        elif shift < 0:
            conv_stop += shift
        # padding
        d = delta * x_pad
        # pre-
        for i in range(seg_start, conv_start):
            y_error[i] -= d
        # post-
        for i in range(conv_stop, seg_stop):
            y_error[i] -= d
        # part of the segment that is affected
        for i in range(conv_start, conv_stop):
            y_error[i] -= delta * x[i - shift]


cdef class Step:
    cdef readonly:
        long i_step, i_stim, i_time
        double delta, e_test, e_train

    def __init__(self, i_step, i_stim, i_time, delta, e_test, e_train):
        self.i_step = i_step
        self.i_stim = i_stim
        self.i_time = i_time
        self.delta = delta
        self.e_test = e_test
        self.e_train = e_train

    def __repr__(self):
        return f"{self.i_step:4}: {self.i_stim:4}, {self.i_time:4} {self.delta:+} --> {self.e_train:10f} / {self.e_test:10f}"


def boosting_fit(
        FLOAT64 [:] y,  # (n_times,)
        FLOAT64 [:,:] x,  # (n_x, n_times)
        FLOAT64 [:] x_pads,  # (n_x,)
        INT64[:,:] split_train,
        INT64[:,:] split_validate,
        INT64[:,:] split_train_and_validate,
        INT64 [:] i_start_by_x,  # (n_x,) kernel start index
        INT64 [:] i_stop_by_x, # (n_x,) kernel stop index
        double delta,
        double mindelta,
        int error,
        int selective_stopping = 0,
):
    """Single model fit using boosting

    Parameters
    ----------
    y : array (n_times,)
        Dependent signal, time series to predict.
    x : array (n_x, n_times)
        Stimulus.
    x_pads : array (n_x,)
        Padding for x.
    split_train
        Training data index.
    split_validate
        Validation data index.
    split_train_and_validate
        Training and validation data index.
    i_start_by_x : ndarray
        Array of i_start for trfs.
    i_stop_by_x : ndarray
        Array of i_stop for TRF.
    delta : scalar
        Step of the adjustment.
    mindelta : scalar
        Smallest delta to use. If no improvement can be found in an iteration,
        the first step is to divide delta in half, but stop if delta becomes
        smaller than ``mindelta``.
    error: int
        Error function to use (1 for l1, 2 for l2).
    selective_stopping : int
        Selective stopping.
    """
    cdef:
        BoostingRunResult *result
        BoostingStep *step
        Step step2
        Py_ssize_t i_start = np.min(i_start_by_x)
        Py_ssize_t n_x = len(x)
        Py_ssize_t n_times_h = np.max(i_stop_by_x) - i_start
        FLOAT64[:, :] h = np.empty((n_x, n_times_h))

    result = boosting_run(y, x, x_pads, h, split_train, split_validate, split_train_and_validate, i_start_by_x, i_stop_by_x, delta, mindelta, error, selective_stopping, i_start, n_times_h)
    out = []
    step = result.history
    while step != NULL:
        step2 = Step(step.i_step, step.i_stim, step.i_time, step.delta, step.e_test, step.e_train)
        out.insert(0, step2)
        step = step.previous
    free_history(result)
    return np.asarray(h), out



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Covariance matrix computation"""
from dataclasses import dataclass, field
from pathlib import Path

import mne
import numpy


@dataclass
class RawCovariance:
    session: str = None
    method: str = 'empirical'
    key: str = field(init=False, default=None)

    def make(
            self,
            raw: mne.io.BaseRaw,
    ) -> mne.Covariance:
        if self.method == 'ad_hoc':
            return mne.cov.make_ad_hoc_cov(raw.info)
        return mne.compute_raw_covariance(raw, method=self.method)


@dataclass
class EpochCovariance:
    epoch: str
    method: str = 'empirical'
    keep_sample_mean: bool = True
    key: str = field(init=False, default=None)

    def make(
            self,
            epochs: mne.Epochs,
            log_path: str,
    ) -> mne.Covariance:
        method = 'empirical' if self.method == 'best' else self.method
        cov = mne.compute_covariance(epochs, self.keep_sample_mean, method=method)

        if self.method == 'best':
            if mne.pick_types(epochs.info, meg='grad', eeg=True, ref_meg=False).size:
                raise NotImplementedError(f"cov={self.key!r}: 'best' regularization is not implemented for EEG or gradiometer sensors; use a different setting for cov.")
            elif epochs is None:
                raise NotImplementedError(f"cov={self.key!r}: 'best' regularization is not implemented for covariance based on raw data; use a different setting for cov.")
            reg_vs = numpy.arange(0, 0.21, 0.01)
            covs = [mne.cov.regularize(cov, epochs.info, mag=v, rank=None) for v in reg_vs]

            # compute whitened global field power
            evoked = epochs.average()
            picks = mne.pick_types(evoked.info, meg='mag', ref_meg=False)
            gfps = [mne.whiten_evoked(evoked, cov, picks).data.std(0) for cov in covs]
            vs = [gfp.mean() for gfp in gfps]
            i = numpy.argmin(numpy.abs(1 - numpy.array(vs)))
            cov = covs[i]
            values = '\n'.join([f"{reg:.2f}: {gfp}" for reg, gfp in zip(reg_vs, gfps)])
            Path(log_path).write_text(f'Picked mag={reg_vs[i]}\nGFP:\n{values}')

        return cov



"""WxPython-based implementation of the Eelbrain ui functions."""
from ..._wxgui import wx, get_app


def ask_saveas(title, message, filetypes, defaultDir, defaultFile):
    """See eelbrain.ui documentation"""
    app = get_app()
    return app.ask_saveas(title, message, filetypes, defaultDir, defaultFile)


def ask_dir(title="Select Folder", message="Please Pick a Folder", must_exist=True):
    app = get_app()
    return app.ask_for_dir(title, message, must_exist)


def ask_file(title, message, filetypes, directory, mult):
    app = get_app()
    return app.ask_for_file(title, message, filetypes, directory, mult)


def ask(title="Overwrite File?",
        message="Duplicate filename. Do you want to overwrite?",
        cancel=False,
        default=True,  # True=YES, False=NO, None=Nothing
        ):
    style = wx.YES_NO | wx.ICON_QUESTION
    if cancel:
        style = style | wx.CANCEL
    if default:
        style = style | wx.YES_DEFAULT
    elif default == False:
        style = style | wx.NO_DEFAULT
    dialog = wx.MessageDialog(None, message, title, style)
    answer = dialog.ShowModal()
    if answer == wx.ID_NO:
        return False
    elif answer == wx.ID_YES:
        return True
    elif answer == wx.ID_CANCEL:
        return None


def ask_color(default=(0, 0, 0)):
    dlg = wx.ColourDialog(None)
    dlg.GetColourData().SetChooseFull(True)
    if dlg.ShowModal() == wx.ID_OK:
        data = dlg.GetColourData()
        out = data.GetColour().Get()
        out = tuple([o / 255. for o in out])
    else:
        out = False
    dlg.Destroy()
    return out


def ask_str(message, title, default=''):
    app = get_app()
    return app.ask_for_string(title, message, default)


def message(title, message="", icon='i'):
    style = wx.OK
    if icon == 'i':
        style = style | wx.ICON_INFORMATION
    elif icon == '?':
        style = style | wx.ICON_QUESTION
    elif icon == '!':
        style = style | wx.ICON_EXCLAMATION
    elif icon == 'error':
        style = style | wx.ICON_ERROR
    elif icon is None:
        pass
    else:
        raise ValueError("Invalid icon argument: %r" % icon)
    dlg = wx.MessageDialog(None, message, title, style)
    dlg.ShowModal()


def copy_file(path):
    if wx.TheClipboard.Open():
        try:
            data_object = wx.FileDataObject()
            data_object.AddFile(path)
            wx.TheClipboard.SetData(data_object)
        except:
            wx.TheClipboard.Close()
            raise
        else:
            wx.TheClipboard.Close()


def copy_text(text):
    if wx.TheClipboard.Open():
        try:
            data_object = wx.TextDataObject(text)
            wx.TheClipboard.SetData(data_object)
        except:
            wx.TheClipboard.Close()
            raise
        else:
            wx.TheClipboard.Close()



"""Data-specific colormaps"""
from itertools import cycle
import logging
from math import ceil
from numbers import Real
from typing import Sequence, List, Tuple, Union

# colormath starts out at 0; needs to be set before init
logger = logging.getLogger('colormath.color_conversions')
if logger.level == 0:  # otherwise it was probably set by user (DEBUG=10)
    logger.setLevel(logging.WARNING)

from colormath.color_objects import LCHabColor, sRGBColor
from colormath.color_conversions import convert_color
import matplotlib
from matplotlib.colors import ListedColormap
from matplotlib.colors import LinearSegmentedColormap, to_rgb, to_rgba
import numpy as np


# https://jfly.uni-koeln.de/html/color_blind/ (Fig. 16)
UNAMBIGUOUS_COLORS = {
    'black': (0.00, 0.00, 0.00),
    'orange': (0.90, 0.60, 0.00),
    'sky blue': (0.35, 0.70, 0.90),
    'bluish green': (0.00, 0.60, 0.50),
    'yellow': (0.95, 0.90, 0.25),
    'blue': (0.00, 0.45, 0.70),
    'vermilion': (0.80, 0.40, 0.00),
    'reddish purple': (0.80, 0.60, 0.70),
}

# How to handle different colormaps
SYMMETRIC_CMAPS = []
ZEROBASED_CMAPS = []
ALPHA_CMAPS = {}  # corresponding cmaps with transparency (alpha channel)

def register_cmap(
        cmap: matplotlib.colors.Colormap,
        symmetric: bool = False,
        zero_based: bool = False,
        alpha_version: Union[bool, str] = None,
):
    matplotlib.colormaps.register(cmap)
    if symmetric:
        SYMMETRIC_CMAPS.append(cmap.name)
    if zero_based:
        ZEROBASED_CMAPS.append(cmap.name)
    if alpha_version is True:
        ALPHA_CMAPS[cmap.name] = f'{cmap.name}-a'
    elif alpha_version:
        ALPHA_CMAPS[cmap.name] = alpha_version


class LocatedColormap:
    vmin = None
    vmax = None
    symmetric = False


class LocatedListedColormap(ListedColormap, LocatedColormap):
    pass


class LocatedLinearSegmentedColormap(LinearSegmentedColormap, LocatedColormap):
    pass


def lch_to_rgb(lightness, chroma, hue):
    "Convert Lightness/Chroma/Hue color representation to RGB"
    psych = LCHabColor(lightness, chroma, hue * 360)
    rgb = convert_color(psych, sRGBColor)
    return rgb.clamped_rgb_r, rgb.clamped_rgb_g, rgb.clamped_rgb_b


def rgb_to_lch(r, g, b):
    rgb = sRGBColor(r, g, b)
    lch = convert_color(rgb, LCHabColor)
    return lch.lch_l, lch.lch_c, lch.lch_h / 360


def _to_rgb(arg, alpha=False):
    if isinstance(arg, (float, int)):
        arg = lch_to_rgb(50, 100, arg)

    if alpha:
        return to_rgba(arg)
    else:
        return to_rgb(arg)


def make_seq_cmap(seq, val, name):
    """Colormap from sequence of RGB values

    Parameters
    ----------
    seq : iterator
        Each entry is either an RGB tuple (if pre- and post color are
        identical) or a tuple with two RGB tuples (separate pre- and post-
        colors).
    val : iterator
        For each entry in ``seq``, the coordinate on the colormap.
    name : str
        Colormap name.
    """
    red = []
    green = []
    blue = []
    for v, col in zip(val, seq):
        if len(col) == 3:
            r0, g0, b0 = col
            r1, g1, b1 = col
        elif len(col) == 2:
            (r0, g0, b0), (r1, g1, b1) = col
        else:
            raise ValueError('col in seq: %s' % str(col))
        red.append((v, r0, r1))
        green.append((v, g0, g1))
        blue.append((v, b0, b1))
    cdict = {'red': red, 'green': green, 'blue': blue}
    return LinearSegmentedColormap(name, cdict)


def unambiguous_color(color: str, lightness: float = None, chroma: float = None):
    """Generate an `unambiguous color <https://jfly.uni-koeln.de/html/color_blind/#pallet>`_
    
    Parameters
    ----------
    color
        Color name, one of: black, orange, sky blue, bluish green, yellow,
        blue, vermilion, reddish purple.
    lightness
        Adjust the lightness (in percent).
    chroma
        Adjust the saturation (in percent).
    """
    r, g, b = UNAMBIGUOUS_COLORS[color]
    l, c, h = rgb_to_lch(r, g, b)
    if lightness is not None:
        l = lightness
    if chroma is not None:
        c = chroma
    return lch_to_rgb(l, c, h)


def twoway_cmap(n1, hue_start=0.1, hue_shift=0.5, name=None, hues=None):
    """Create colormap for two-way interaction

    Parameters
    ----------
    n1 : int
        Number of levels on the first factor.
    hue_start : 0 <= scalar < 1
        First hue value.
    hue_shift : 0 <= scalar < 1
        Use that part of the hue continuum between categories to shift hue
        within categories.
    name : str
        Name of the colormap.
    hues : list of scalar
        List of hue values corresponding to the levels of the first factor
        (overrides regular hue distribution).
    """
    # within each hue, create values for [-1, -0.5, 0.5, 1]
    # return list of [i, (r, g, b), (r, g, b)]
    if hues is None:
        hues = np.linspace(hue_start, hue_start + 1, n1, False) % 1.
    hue_shift *= (0.5 / n1)

    seqs = []
    for h in hues:
        h_pre = (h - hue_shift) % 1
        h_post = (h + hue_shift) % 1
        seqs.append((lch_to_rgb(0, 100, h_pre),
                     lch_to_rgb(50, 100, h),
                     lch_to_rgb(100, 100, h_post)))

    seq = []
    for i in range(n1):
        seq.append((seqs[i - 1][-1], seqs[i][0]))
        seq.append(seqs[i][1])
        if i == n1 - 1:
            seq.append((seqs[i][2], seqs[0][0]))

    loc = np.linspace(0, 1, n1 * 2 + 1)

    if name is None:
        name = "%i_by_n" % n1

    return make_seq_cmap(seq, loc, name)


def oneway_colors(
        n: int,
        hue_start: Union[float, Sequence[float]] = 0.2,
        light_range: Union[float, Tuple[float, float]] = 0.5,
        light_cycle: int = None,
        always_cycle_hue: bool = False,
        locations: Sequence[float] = None,
        unambiguous: Union[bool, Sequence[int]] = None,
) -> list:
    "Create colors for categories (see docs at colors_for_oneway)"
    if unambiguous:
        if unambiguous is True:
            if n > 8:
                raise ValueError("unambiguous=True for n > 8")
            indices = list(range(n))
        else:
            if min(unambiguous) < 1 or max(unambiguous) > 8:
                raise ValueError(f"unambiguous={unambiguous}: values outside range (1, 8)")
            indices = [i - 1 for i in unambiguous]
        colors = list(UNAMBIGUOUS_COLORS.values())
        return [colors[i] for i in indices]

    if locations is not None:
        raise NotImplementedError('locations for non-cmap based colors')

    if light_cycle is None or always_cycle_hue:
        n_hues = n
    else:
        n_hues = int(ceil(n / light_cycle))

    if isinstance(hue_start, Real):
        hue = np.linspace(hue_start, hue_start + 1, n_hues, False) % 1.
        if light_cycle and not always_cycle_hue:
            hue = np.repeat(hue, light_cycle)
    elif len(hue_start) >= n_hues:
        hue = hue_start
    else:
        raise ValueError("If list of hues is provided it needs ot contain at least as many hues as there are cells")

    if isinstance(light_range, (list, tuple)):
        start, stop = light_range
    else:
        start = 0.5 + 0.5 * light_range
        stop = 0.5 - 0.5 * light_range

    if light_cycle is None:
        lightness = np.linspace(100 * start, 100 * stop, n)
    else:
        tile = np.linspace(100 * start, 100 * stop, light_cycle)
        lightness = cycle(tile)
        if n % light_cycle:
            hue = hue[:n]

    return [lch_to_rgb(l, 100, h) for l, h in zip(lightness, hue)]


def twoway_colors(
        n1: int,
        n2: int,
        hue_start: float = 0.2,
        hue_shift: float = 0.,
        hues: Sequence[float] = None,
        lightness: Union[float, Sequence[float]] = None,
) -> List[Tuple[float, float, float]]:
    """Create colors for two-way interaction

    Parameters
    ----------
    n1
        Number of levels on the first factor.
    n2
        Number of levels on the second factor.
    hue_start : 0 <= scalar < 1
        First hue value.
    hue_shift : 0 <= scalar < 1
        Use that part of the hue continuum between categories to shift hue
        within categories.
    hues
        List of hue values corresponding to the levels of the first factor
        (overrides regular hue distribution).
    lightness
        If specified as scalar, colors will occupy the range
        ``[lightness, 100-lightness]``. Can also be given as list with one
        value corresponding to each element in the second factor.
    """
    if hues is None:
        hues = np.linspace(hue_start, hue_start + 1, n1, False) % 1.
    else:
        hues = np.asarray(hues)
        if np.any(hues > 1) or np.any(hues < 0):
            raise ValueError(f"hues={hues}: values out of range, need to be in [0, 1]")
        elif len(hues) < n1:
            raise ValueError(f"hues={hues}: need as many hues as levels in the first factor (got {len(hues)}, need {n1})")
    hue_shift *= (1. / 3. / n1)

    if lightness is None:
        lstart = 60. / n2
        ls = np.linspace(lstart, 100 - lstart, n2)
    elif isinstance(lightness, Real):
        ls = np.linspace(lightness, 100 - lightness, n2)
    else:
        if len(lightness) != n2:
            raise ValueError(f"lightness={lightness!r}: need {n2} values")
        ls = lightness

    colors = []
    for hue in hues:
        hs = np.linspace(hue - hue_shift, hue + hue_shift, n2) % 1
        colors.extend(lch_to_rgb(l, 100, h) for l, h in zip(ls, hs))

    return colors


def two_step_colormap(left_max, left, center='transparent', right=None, right_max=None, name='two-step'):
    """Colormap with 2 or 4 gradients (e.g., ``white - red - transparent - blue - white``)

    Parameters
    ----------
    left_max : matplotlib color
        Left end of the colormap.
    left : matplotlib color
        Left middle of the colormap.
    center : matplotlib color | 'transparent'
        Color for the middle value; 'transparent to make the middle transparent
        (default).
    right : matplotlib color
        Right middle of the colormap (if not specified, the colormap ends at
        the location specified by ``center``).
    right_max : matplotlib color
        Right end of the colormap.
    name : str
        Name for the colormap.

    Examples
    --------
    Standard red/blue::

        >>> cmap = plot.two_step_colormap('black', 'red', 'transparent', 'blue', 'black', name='red-blue')
        >>> plot.ColorBar(cmap, 1)

    Or somewhat more adventurous::

        >>> cmap = plot.two_step_colormap('black', (1, 0, 0.3), 'transparent', (0.3, 0, 1), 'black', name='red-blue-2')
    """
    if center == 'transparent':
        center_ = None
        transparent_middle = True
    else:
        center_ = _to_rgb(center, False)
        transparent_middle = False
    left_max_ = _to_rgb(left_max, transparent_middle)
    left_ = _to_rgb(left, transparent_middle)
    is_symmetric = right is not None
    if is_symmetric:
        right_ = _to_rgb(right, transparent_middle)
        right_max_ = _to_rgb(right_max, transparent_middle)
    else:
        right_ = right_max_ = None

    kind = (is_symmetric, transparent_middle)
    if kind == (False, False):
        clist = (
            (0.0, center_),
            (0.5, left_),
            (1.0, left_max_),
        )
    elif kind == (False, True):
        clist = (
            (0.0, (*left_[:3], 0)),
            (0.5, left_),
            (1.0, left_max_),
        )
    elif kind == (True, False):
        clist = (
            (0.0, left_max_),
            (0.25, left_),
            (0.5, center_),
            (0.75, right_),
            (1.0, right_max_),
        )
    elif kind == (True, True):
        clist = (
            (0.0, left_max_),
            (0.25, left_),
            (0.5, (*left_[:3], 0)),
            (0.5, (*right_[:3], 0)),
            (0.75, right_),
            (1.0, right_max_),
        )
    else:
        raise RuntimeError
    cmap = LocatedLinearSegmentedColormap.from_list(name, clist)
    cmap.set_bad('w', alpha=0.)
    cmap.symmetric = is_symmetric
    return cmap


def pigtailed_cmap(cmap, swap_order=('green', 'red', 'blue')):
    # nilearn colormaps with neutral middle
    orig = matplotlib.colormaps.get_cmap(cmap)._segmentdata
    cdict = {
        'green': [(0.5 * (1 - p), *c) for p, *c in reversed(orig[swap_order[0]])],
        'blue': [(0.5 * (1 - p), *c) for p, *c in reversed(orig[swap_order[1]])],
        'red': [(0.5 * (1 - p), *c) for p, *c in reversed(orig[swap_order[2]])],
    }
    for color in ('red', 'green', 'blue'):
        cdict[color].extend((0.5 * (1 + p), *c) for p, *c in orig[color])
    return cdict


def make_cmaps():
    """Create custom colormaps and register them with matplotlib"""
    # Polar
    # -----
    # bi-polar, blue-white-red based
    cmap = LinearSegmentedColormap.from_list(
        "polar", (
            (0.0, (0.0, 0.0, 1.0)),
            (0.5, (1.0, 1.0, 1.0)),
            (1.0, (1.0, 0.0, 0.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True, alpha_version=True)
    # polar-alpha: middle is transparent instead of white
    cmap = LinearSegmentedColormap.from_list(
        "polar-a", (
            (0.0, (0.0, 0.0, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 1.0, 0.0)),
            (0.5, (1.0, 0.0, 0.0, 0.0)),
            (1.0, (1.0, 0.0, 0.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True)

    # xpolar ("extra-polar"): fade ends into black
    cmap = LinearSegmentedColormap.from_list(
        "xpolar", (
            (0.0, (0.0, 0.0, 0.0)),
            (0.3, (0.0, 0.0, 1.0)),
            (0.5, (1.0, 1.0, 1.0)),
            (0.7, (1.0, 0.0, 0.0)),
            (1.0, (0.0, 0.0, 0.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True, alpha_version=True)
    # extra-polar alpha: middle is transparent instead of white
    cmap = LinearSegmentedColormap.from_list(
        "xpolar-a", (
            (0.0, (0.0, 0.0, 0.0, 1.0)),
            (0.3, (0.0, 0.0, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 1.0, 0.0)),
            (0.5, (1.0, 0.0, 0.0, 0.0)),
            (0.7, (1.0, 0.0, 0.0, 1.0)),
            (1.0, (0.0, 0.0, 0.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True)

    # extra-polar light: ends are light instead of dark
    cmap = LinearSegmentedColormap.from_list(
        "lpolar", (
            (0.0, (0.5, 1.0, 1.0, 1.0)),
            (0.1, (0.0, 1.0, 1.0, 1.0)),
            (0.3, (0.0, 0.0, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 0.0, 1.0)),
            (0.5, (0.0, 0.0, 0.0, 1.0)),
            (0.7, (1.0, 0.0, 0.0, 1.0)),
            (0.9, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 0.5, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True, alpha_version=True)
    # -alpha
    cmap = LinearSegmentedColormap.from_list(
        "lpolar-a", (
            (0.0, (0.5, 1.0, 1.0, 1.0)),
            (0.1, (0.0, 1.0, 1.0, 1.0)),
            (0.3, (0.0, 0.0, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 1.0, 0.0)),
            (0.5, (1.0, 0.0, 0.0, 0.0)),
            (0.7, (1.0, 0.0, 0.0, 1.0)),
            (0.9, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 0.5, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True)
    # version with alpha higher up
    cmap = LinearSegmentedColormap.from_list(
        "lpolar-aa", (
            (0.0, (0.5, 1.0, 1.0, 1.0)),
            (0.1, (0.0, 1.0, 1.0, 1.0)),
            (0.3, (0.0, 0.0, 1.0, 0.5)),
            (0.5, (0.0, 0.0, 1.0, 0.0)),
            (0.5, (1.0, 0.0, 0.0, 0.0)),
            (0.7, (1.0, 0.0, 0.0, 0.5)),
            (0.9, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 0.5, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True)

    # Lux
    # ---
    # lux:  cyan vs orange
    cmap = LinearSegmentedColormap.from_list(
        "lux", (
            (0.0, (1.0, 1.0, 1.0, 1.0)),
            (0.15, (0.0, 1.0, 1.0, 1.0)),
            (0.4, (0.0, 0.3, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 0.0, 1.0)),
            (0.5, (0.0, 0.0, 0.0, 1.0)),
            (0.6, (1.0, 0.3, 0.0, 1.0)),
            (0.85, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 1.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True, alpha_version=True)
    # lux-gray:  gray center
    cmap = LinearSegmentedColormap.from_list(
        "lux-gray", (
            (0.0, (1.0, 1.0, 1.0, 1.0)),
            (0.15, (0.0, 1.0, 1.0, 1.0)),
            (0.4, (0.0, 0.3, 1.0, 1.0)),
            (0.5, (0.9, 0.9, 0.9, 1.0)),
            (0.6, (1.0, 0.3, 0.0, 1.0)),
            (0.85, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 1.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True, alpha_version='lux-a')
    # -alpha
    cmap = LinearSegmentedColormap.from_list(
        "lux-a", (
            (0.0, (1.0, 1.0, 1.0, 1.0)),
            (0.15, (0.0, 1.0, 1.0, 1.0)),
            (0.4, (0.0, 0.3, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 1.0, 0.0)),
            (0.5, (1.0, 0.0, 0.0, 0.0)),
            (0.6, (1.0, 0.3, 0.0, 1.0)),
            (0.85, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 1.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True)

    # polar-lux:  polar on the inside, blend to lux
    cmap = LinearSegmentedColormap.from_list(
        "polar-lux", (
            (0.0, (1.0, 1.0, 1.0, 1.0)),
            (0.1, (0.0, 1.0, 1.0, 1.0)),
            (0.4, (0.0, 0.0, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 0.0, 1.0)),
            (0.5, (0.0, 0.0, 0.0, 1.0)),
            (0.6, (1.0, 0.0, 0.0, 1.0)),
            (0.9, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 1.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True, alpha_version=True)
    # -alpha
    cmap = LinearSegmentedColormap.from_list(
        "polar-lux-a", (
            (0.0, (1.0, 1.0, 1.0, 1.0)),
            (0.1, (0.0, 1.0, 1.0, 1.0)),
            (0.4, (0.0, 0.0, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 1.0, 0.0)),
            (0.5, (1.0, 0.0, 0.0, 0.0)),
            (0.6, (1.0, 0.0, 0.0, 1.0)),
            (0.9, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 1.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True)

    # lux-purple:  on the negative side, blend into purple instead of cyan
    cmap = LinearSegmentedColormap.from_list(
        "lux-purple", (
            (0.0, (1.0, 1.0, 1.0, 1.0)),
            (0.15, (1.0, 0.0, 1.0, 1.0)),
            (0.4, (0.3, 0.0, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 0.0, 1.0)),
            (0.5, (0.0, 0.0, 0.0, 1.0)),
            (0.6, (1.0, 0.3, 0.0, 1.0)),
            (0.85, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 1.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True, alpha_version=True)
    cmap = LinearSegmentedColormap.from_list(
        "lux-purple-a", (
            (0.0, (1.0, 1.0, 1.0, 1.0)),
            (0.15, (1.0, 0.0, 1.0, 1.0)),
            (0.4, (0.3, 0.0, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 1.0, 0.0)),
            (0.5, (1.0, 0.0, 0.0, 0.0)),
            (0.6, (1.0, 0.3, 0.0, 1.0)),
            (0.85, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 1.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True)

    cmap = LinearSegmentedColormap.from_list(
        "polar-lux-purple", (
            (0.0, (1.0, 1.0, 1.0, 1.0)),
            (0.1, (1.0, 0.0, 1.0, 1.0)),
            (0.4, (0.0, 0.0, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 0.0, 1.0)),
            (0.5, (0.0, 0.0, 0.0, 1.0)),
            (0.6, (1.0, 0.0, 0.0, 1.0)),
            (0.9, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 1.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True, alpha_version=True)
    cmap = LinearSegmentedColormap.from_list(
        "polar-lux-purple-a", (
            (0.0, (1.0, 1.0, 1.0, 1.0)),
            (0.1, (1.0, 0.0, 1.0, 1.0)),
            (0.4, (0.0, 0.0, 1.0, 1.0)),
            (0.5, (0.0, 0.0, 1.0, 0.0)),
            (0.5, (1.0, 0.0, 0.0, 0.0)),
            (0.6, (1.0, 0.0, 0.0, 1.0)),
            (0.9, (1.0, 1.0, 0.0, 1.0)),
            (1.0, (1.0, 1.0, 1.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap, symmetric=True)

    # phase
    # -----
    cmap = LinearSegmentedColormap.from_list(
        "phase", (
            (0.0, (0.0, 0.0, 1.0)),
            (0.5, (1.0, 0.0, 0.0)),
            (1.0, (0.0, 0.0, 1.0)),
        ))
    cmap.set_bad('w', alpha=0.)
    register_cmap(cmap)

    # Significance
    # ------------
    # sig: significance map for specific vmax=0.05
    cmap = LinearSegmentedColormap.from_list(
        "sig", (
            (0.0,  (1.0, 1.0, 1.0)),
            (0.02, (1.0, 1.0, 0.0)),
            (0.2,  (1.0, 0.5, 0.0)),
            (1.0,  (1.0, 0.0, 0.0)),
        ))
    cmap.set_over('k', alpha=0.)
    cmap.set_bad('b', alpha=0.)
    register_cmap(cmap, zero_based=True)

    # Nilearn cmaps
    if 'cold_hot' not in matplotlib.colormaps:
        cmap = LinearSegmentedColormap('cold_hot', pigtailed_cmap('hot'))
        register_cmap(cmap, symmetric=True)
    if 'cold_white_hot' not in matplotlib.colormaps:
        cmap = LinearSegmentedColormap('cold_white_hot', pigtailed_cmap('hot_r'))
        register_cmap(cmap, symmetric=True)


make_cmaps()

# https://matplotlib.org/tutorials/colors/colormaps.html
mpl_diverging = ('PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu', 'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic')
SYMMETRIC_CMAPS.extend(mpl_diverging)
SYMMETRIC_CMAPS.extend(f'{name}_r' for name in mpl_diverging)



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
"""Help Viewer"""
import wx

from ..fmtxt import make_html_doc, Section, Code
from .text import HTMLFrame


def show_help_txt(text, parent, title=""):
    "Show help frame with text in mono-spaced font"
    s = Section(title, Code(text))
    html = make_html_doc(s, None)
    frame = HelpFrame(parent, title, html)


class HelpFrame(HTMLFrame):

    def __init__(self, parent, title, html, **kwargs):
        display_w, display_h = wx.DisplaySize()
        x = 0
        y = 25
        w = min(700, display_w)
        h = min(2000, display_h - y - 100)
        HTMLFrame.__init__(self, parent, title, html, pos=(x, y), size=(w, h), **kwargs)



from ._experiment.mne_experiment import MneExperiment
from ._experiment.preprocessing import RawSource, RawFilter, RawICA, RawMaxwell, RawOversampledTemporalProjection, RawReReference, RawApplyICA
from ._experiment.epochs import ContinuousEpoch, EpochCollection, PrimaryEpoch, SecondaryEpoch, SuperEpoch
from ._experiment.groups import Group, SubGroup
from ._experiment.parc import SubParc, CombinationParc, FreeSurferParc, FSAverageParc, SeededParc, IndividualSeededParc
from ._experiment.test_def import ANOVA, TTestOneSample, TTestIndependent, TTestRelated, TContrastRelated, TwoStageTest, ROITestResult, ROI2StageResult
from ._experiment.variable_def import EvalVar, GroupVar, LabelVar



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from pathlib import Path

from eelbrain import datasets, load, save
from eelbrain.testing import TempDir, requires_pyarrow, assert_dataset_equal


@requires_pyarrow
def test_arrow_io():
    tempdir = TempDir()
    ds = datasets.get_uts()

    path = Path(tempdir) / 'test.arrow'
    save.arrow(ds, path)
    ds2 = load.arrow(path)
    assert_dataset_equal(ds2, ds)



"""Helper functions for dealing with logging

By default, mne's logger has a highly verbose stream handler logging to the
screen. Here this screen handler is set to a higher level (WARNING), but only
if Eelbrain is the primary import (i.e., it is imported before mne) and if the
user has not stored another configuration.
"""
# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from logging import WARNING, FileHandler, Formatter, StreamHandler, getLogger
import sys
first_mne_import = 'mne' not in sys.modules

import mne

from .._utils import log_level


def reset_logger(logger):
    # lower mne's screen logging level, but defer to user-defined setting
    level = log_level(mne.get_config('MNE_LOGGING_LEVEL', WARNING))
    formatter = Formatter("%(levelname)-8s %(name)s:%(message)s")
    for h in logger.handlers:
        if isinstance(h, StreamHandler):
            h.setFormatter(formatter)
            if h.level < level:
                h.setLevel(level)


if first_mne_import:
    reset_logger(mne.utils.logger)


class CaptureLog:
    """Context to capture log from a specific logger and write it to a file

    Parameters
    ----------
    filename : str
        Where to write the log file.
    mode : str
        Mode for opening the log file (default 'w').
    name : str
        Name of the logger from which to capture (default 'mne').
    """
    def __init__(self, filename, mode='w', logger='mne', level='debug'):
        self.logger = logger
        self.level = log_level(level)
        self.handler = FileHandler(filename, mode)
        self.handler.setLevel(self.level)
        self.handler.setFormatter(Formatter("%(levelname)-8s :%(message)s"))
        self._old_level = None

    def __enter__(self):
        logger = getLogger(self.logger)
        logger.addHandler(self.handler)
        if logger.level == 0 or logger.level > self.level:
            self._old_level = logger.level
            logger.setLevel(self.level)
        return logger

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.handler.close()
        logger = getLogger(self.logger)
        logger.removeHandler(self.handler)
        if self._old_level is not None:
            logger.setLevel(self._old_level)



"""Basic GUI methods using Tkinter

http://tkinter.unpythonic.net/wiki/tkFileDialog
"""
from tkinter import Tk
import tkinter.filedialog
from tkinter.filedialog import Open, SaveAs
import tkinter.messagebox
import tkinter.simpledialog


root = Tk()  # initialize application
root.withdraw()  # hide root window


def ask_saveas(title, message, filetypes, defaultDir=None, defaultFile=None):
    dlg = SaveAs(title=title, message=message, filetypes=filetypes)
    filename = dlg.show()
    return filename


def ask_dir(title="Select Folder", message="Please Pick a Folder",
            must_exist=True):
    return tkinter.filedialog.askdirectory(title=title, mustexist=must_exist)


def ask_file(title, message, filetypes, directory, mult):
    dlg = Open(title=title, filetypes=filetypes, initialdir=directory,
               multiple=mult)
    out = dlg.show()
    return out


def ask(title="Overwrite File?",
        message="Duplicate filename. Do you want to overwrite?",
        cancel=False, default=True,  # True=YES, False=NO, None=Nothing
        ):
    return tkinter.messagebox.askyesno(title, message)


def ask_str(message, title, default=''):
    return tkinter.simpledialog.askstring(title, message, initialvalue=default)


def copy_file(path):
    raise NotImplementedError


def copy_text(text):
    # http://stackoverflow.com/a/4203897/166700
    root.withdraw()
    root.clipboard_clear()
    root.clipboard_append(text)
    root.destroy()


def message(title, message="", icon='i'):
    if icon in 'i?':
        tkinter.messagebox.showinfo(title, message)
    elif icon == '!':
        tkinter.messagebox.showwarning(title, message)
    elif icon == 'error':
        tkinter.messagebox.showerror(title, message)
    else:
        raise ValueError("Invalid icon argument: %r" % icon)


def show_help(obj):
    print(getattr(obj, '__doc__', 'no docstring'))



# Author: Christian Brodbeck <christianbrodbeck@nyu.edu>
from ._sensor import sensor_results, sensor_time_results, sensor_time_frequency_results
from ._source import source_results, source_time_lm, source_time_results
from ._uts import time_results
from ._uv import scatter_table
